from pydantic import BaseModel, ConfigDict


class TestClass(BaseModel):
    model_config = ConfigDict(frozen=True)

    a: int
    b: int
    s: str
    x: float
    y: float

    def get_csv_header(self):
        return "a,b,s,x,y"

    def to_csv_row(self):
        return f"{self.a},{self.b},{self.s},{self.x},{self.y}"


test_records = [
    TestClass(a=1, b=2, s="hello", x=3.14, y=2.71),
    TestClass(a=2, b=3, s="world", x=2.71, y=3.14),
]



from collections.abc import Callable, Iterable
from pathlib import Path
from typing import Literal, TextIO, TypeAlias

from pydantic import BaseModel, RootModel

from . import csv as csv_backend
from . import json as json_backend
from . import json_lines as jsl_backend
from .common import T
from .version import __version__

try:
    from . import yaml as yaml_backend
except ImportError:
    from . import yaml_stub as yaml_backend

GenericDataFormat = Literal["json", "yaml"]
LinesOnlyDataFormat = Literal["csv", "json_lines"]
DataFormat = GenericDataFormat | LinesOnlyDataFormat


def decide_data_format_from_path(
    file_path: Path,
) -> DataFormat:
    match file_path.suffix.lower():
        case ".csv":
            return "csv"
        case ".jsonl" | ".jsl" | ".jl" | ".json_lines":
            return "json_lines"
        case ".json":
            return "json"
        case ".yaml" | ".yml":
            return "yaml"
        case _:
            raise ValueError(f"Unsupported file extension: {file_path.suffix}")


def read_record_from_reader(
    reader: TextIO, model: type[T], data_format: GenericDataFormat
) -> T:
    match data_format:
        case "json":
            return json_backend.read_record(reader, model)
        case "yaml":
            return yaml_backend.read_record(reader, model)
        case _:
            raise ValueError(f"Unsupported backend type: {data_format}")


def read_record_from_file(file_path: str | Path, model: type[T]) -> T:
    file_path = Path(file_path)
    data_format: GenericDataFormat = decide_data_format_from_path(file_path)  # type: ignore
    with file_path.open() as reader:
        return read_record_from_reader(reader, model, data_format)


def read_records_from_reader(
    reader: TextIO,
    model: type[T],
    data_format: DataFormat,
) -> list[T]:
    list_model = RootModel[list[model]]
    match data_format:
        case "csv":
            return csv_backend.read_records(reader, model)
        case "json_lines":
            return jsl_backend.read_records(reader, model)
        case "json":
            return json_backend.read_record(reader, list_model).root
        case "yaml":
            return yaml_backend.read_record(reader, list_model).root
        case _:
            raise ValueError(f"Unsupported backend type: {data_format}")


def read_records_from_file(file_path: str | Path, model: type[T]) -> list[T]:
    file_path = Path(file_path)
    data_format = decide_data_format_from_path(file_path)
    with file_path.open() as reader:
        return read_records_from_reader(reader, model, data_format)


def write_record_to_writer(
    writer: TextIO, record: BaseModel, data_format: GenericDataFormat
) -> None:
    match data_format:
        case "json":
            json_backend.write_record(writer, record)
        case "yaml":
            yaml_backend.write_record(writer, record)
        case _:
            raise ValueError(f"Unsupported backend type: {data_format}")


def write_record_to_file(file_path: str | Path, record: BaseModel) -> None:
    file_path = Path(file_path)
    data_format: GenericDataFormat = decide_data_format_from_path(file_path)  # type: ignore
    with file_path.open("w") as writer:
        write_record_to_writer(writer, record, data_format)


def write_records_to_writer(
    writer: TextIO,
    records: Iterable[T],
    data_format: DataFormat,
) -> None:
    list_model = RootModel[Iterable[T]]
    match data_format:
        case "csv":
            csv_backend.write_records(writer, records)
        case "json_lines":
            jsl_backend.write_records(writer, records)
        case "json":
            json_backend.write_record(writer, list_model(root=records))
        case "yaml":
            yaml_backend.write_record(writer, list_model(root=records))
        case _:
            raise ValueError(f"Unsupported backend type: {data_format}")


def write_records_to_file(file_path: str | Path, records: Iterable[T]) -> None:
    file_path = Path(file_path)
    data_format = decide_data_format_from_path(file_path)
    with file_path.open("w") as writer:
        write_records_to_writer(writer, records, data_format)



from collections.abc import Iterable
from typing import TextIO

from .common import T


def read_records(reader: TextIO, model: type[T]) -> list[T]:
    return [model.model_validate_json(line) for line in reader]


def write_records(writer: TextIO, records: Iterable[T]) -> None:
    for record in records:
        writer.write(record.model_dump_json())
        writer.write("\n")



from typing import TextIO

from pydantic import BaseModel

from .common import T


def read_record(reader: TextIO, model: type[T]) -> T:
    raise NotImplementedError("yaml backend is not available.")


def write_record(writer: TextIO, record: BaseModel) -> None:
    raise NotImplementedError("yaml backend is not available.")



from pathlib import Path
from pydanticio import decide_data_format_from_path

from pytest import raises


def test_decide_data_format_from_path():
    assert decide_data_format_from_path(Path("test.csv")) == "csv"
    assert decide_data_format_from_path(Path("test.json")) == "json"
    assert decide_data_format_from_path(Path("test.jsonl")) == "json_lines"

    with raises(ValueError):
        decide_data_format_from_path(Path("test.txt"))

    with raises(ValueError):
        decide_data_format_from_path(Path("test"))

    with raises(ValueError):
        decide_data_format_from_path(Path("test."))



import csv
from collections.abc import Iterable
from typing import TextIO

from .common import T


def read_records(reader: TextIO, model: type[T]) -> list[T]:
    csv_reader = csv.DictReader(reader)
    return [model.model_validate(row) for row in csv_reader]


def write_records(writer: TextIO, records: Iterable[T]) -> None:
    it = iter(records)
    first_record = next(it)
    fields = list(first_record.model_fields.keys())
    csv_writer = csv.DictWriter(writer, fieldnames=fields)
    csv_writer.writeheader()
    csv_writer.writerow(first_record.model_dump(mode="json"))
    for record in it:
        csv_writer.writerow(record.model_dump(mode="json"))



from typing import TextIO

import yaml
from pydantic import BaseModel

from .common import T


def read_record(reader: TextIO, model: type[T]) -> T:
    data = yaml.safe_load(reader)
    return model.model_validate(data)


def write_record(writer: TextIO, record: BaseModel) -> None:
    yaml.safe_dump(record.model_dump(mode="json"), writer)



from typing import TextIO

from pydantic import BaseModel

from .common import T


def read_record(reader: TextIO, model: type[T]) -> T:
    data = reader.read()
    return model.model_validate_json(data)


def write_record(writer: TextIO, record: BaseModel) -> None:
    writer.write(record.model_dump_json())



from pydantic import BaseModel
from pydanticio import read_records_from_file, write_records_to_file


class User(BaseModel):
    name: str
    age: int


def main(input_path: str, output_path: str) -> None:
    users = read_records_from_file(input_path, User)
    write_records_to_file(output_path, users)


if __name__ == "__main__":
    import sys

    main(sys.argv[1], sys.argv[2])



from io import StringIO

from pydanticio import (
    read_record_from_reader,
    write_record_to_writer,
    read_records_from_reader,
    write_records_to_writer,
)
from pydantic import RootModel

from . import TestClass, test_records


def test_read_record_from_reader():
    record_str = test_records[0].model_dump_json()
    reader = StringIO(record_str)
    record = read_record_from_reader(reader, TestClass, "json")
    assert record == test_records[0]


def test_read_list_of_records_from_reader():
    records_str = ",".join(record.model_dump_json() for record in test_records)
    records_str = f"[{records_str}]"
    reader = StringIO(records_str)
    records = read_records_from_reader(reader, TestClass, "json")
    assert records == test_records


def test_write_record_to_writer():
    record_str = test_records[0].model_dump_json()
    writer = StringIO()
    write_record_to_writer(writer, test_records[0], "json")
    assert writer.getvalue() == record_str


def test_write_list_of_records_to_writer():
    records_str = ",".join(record.model_dump_json() for record in test_records)
    records_str = f"[{records_str}]"
    writer = StringIO()
    write_records_to_writer(writer, test_records, "json")
    assert writer.getvalue() == records_str



from io import StringIO

from pydanticio import read_records_from_reader, write_records_to_writer

from . import TestClass, test_records

record_lines = [record.model_dump_json() for record in test_records]


def test_read_records_from_reader():
    reader = StringIO("\n".join(record_lines))
    records = read_records_from_reader(reader, TestClass, "json_lines")
    assert records == test_records


def test_write_records_to_writer():
    writer = StringIO()
    write_records_to_writer(writer, test_records, "json_lines")
    assert writer.getvalue().strip().splitlines() == record_lines



from io import StringIO

from pydanticio import read_records_from_reader, write_records_to_writer

from . import TestClass, test_records


records_str = "\r\n".join(
    [
        test_records[0].get_csv_header(),
        test_records[0].to_csv_row(),
        test_records[1].to_csv_row(),
    ]
)


def test_read_records_from_reader():
    reader = StringIO(records_str)
    records = read_records_from_reader(reader, TestClass, "csv")
    assert records == test_records


def test_write_records_to_writer():
    writer = StringIO()
    write_records_to_writer(writer, test_records, "csv")
    assert writer.getvalue().strip() == records_str.strip()
